#!/bin/bash
#
# kdenlive-install:kdenlive-install.sh
#
# Install last version of kdenlive (using AppImage format) and add menu entry
#
set -u
set -o pipefail

declare -gr DEFAULT_APPIMAGES_INSTALL_COMMON_HOME='/usr/local/share/appimages'
declare -gr DEFAULT_APPIMAGES_INSTALL_ALL_USERS_KDENLIVE_HOME="${DEFAULT_APPIMAGES_INSTALL_COMMON_HOME}/kdenlive"

declare -gr DEFAULT_KDENLIVE_LIST_VERSIONS_URL='https://download.kde.org/stable/kdenlive/'
declare -gr DEFAULT_KDENLIVE_LIST_APPIMAGES_URL_PATTERN='https://download.kde.org/stable/kdenlive/%s/linux/'

declare -gr DEFAULT_KDENLIVE_ICON_URL='http://www.iconarchive.com/download/i105845/papirus-team/papirus-apps/kdenlive.svg'

declare -gr COMMON_APPIMAGE_LINK="${DEFAULT_APPIMAGES_INSTALL_ALL_USERS_KDENLIVE_HOME}/kdenlive.appimage"
declare -gr COMMON_DESKTOP="${DEFAULT_APPIMAGES_INSTALL_ALL_USERS_KDENLIVE_HOME}/kdenlive.desktop"
declare -gr COMMON_MENU_ICON_FILENAME_PREFIX="${DEFAULT_APPIMAGES_INSTALL_ALL_USERS_KDENLIVE_HOME}/kdenlive" # + extension
declare -gr COMMON_MENU_ENTRY='kdenlive (appimage)'

function usage {
  cat <<EOF
Install KDEnlive appimage

Usage:

  $( basename "$0" ) [--list]
  [sudo] $( basename "$0" ) [--install] [--force|--force-download|--force-menu] [--home INSTALLATION_DIR] [--url DOWNLOAD_URL] [--icon ICON_URL]
  $( basename "$0" ) --install-user [--force|--force-download|--force-menu] [--home INSTALLATION_DIR] [--url DOWNLOAD_URL] [--icon ICON_URL]
  $( basename "$0" ) --install-all-users [--force|--force-download|--force-menu] [--home INSTALLATION_DIR] [--url DOWNLOAD_URL] [--icon ICON_URL]

Actions:

  -h,   --help      Display this message and exit.
  -l,   --list      List url of available versions.
  -i,   --install   (default action)
                    Install for current user only or for all users depend of if you use 'sudo' or not
  -U,   --install-user
                    Install for current user only
  -A,   --install-all-users
                    Install for all users (installer will use sudo if needed)

Options:

  -F,   --force     Allow to reinstall existing version of kdenlive
        --force-download
                    Force download even if file already present.
        --force-menu
                    Force menu creation.
        --home DIR  Set installation directory
                    (default value: '$( compute_DEFAULT_APPIMAGES_INSTALL_HOME )')
        --icon URL  Set kdenlive icon URL
                    (default value: '${DEFAULT_KDENLIVE_ICON_URL}')
        --skip-update
                    Skip system update
        --debug
                    Add some traces

EOF
}

#
# P_param : get parameter value, if not exist call usage and return error.
#
# Usage:
#   param_xxx="$( P_param "$@" )" && shift || return $?
#
# Require:
#   - param variable with current parameter name
#   - usage function (use to display usage)
#
function P_param {
  if [ "$#" -gt 0 ] ; then
    echo "$1"
  else
    echofatal "Value required for '${param}'"
    usage >&2
    return 1
  fi
}

function echofatal {
  echo '*** [FATAL]: ' "$@" >&2
}

function echowarn {
  echo '* [WARN]: ' "$@" >&2
}

function echoinfo {
  echo '* [INFO]: ' "$@" >&2
}

function set_install_mode {
  local -r action="$1"

  case "${action}" in
    '--install-user'     ) MODE='user' ;;
    '--install-all-users') MODE='system' ;;

    '--install'|'')
      if [[ $(id -u) -ne 0 ]] ; then
        MODE='user'
      else
        MODE='system'
      fi
    ;;

    *)
      echofatal "Internal error: set_install_mode '${action}'"
      exit 100
  esac

  if [ "${MODE}" = 'system' ] ; then
    SUDO_IF_NEEDED='sudo'
  else
    SUDO_IF_NEEDED=''
  fi
}

function fix_install_mode_if_needed {
  if [ "${MODE}" = 'user' ] && [ -d "${DEFAULT_APPIMAGES_INSTALL_ALL_USERS_KDENLIVE_HOME}" ]; then
    while true; do
      # Getting LC_MESSAGES values into variables
      # shellcheck disable=SC2046 # Intended IFS splitting
      IFS='
' set -- $(locale LC_MESSAGES)

      yesexpr="$1"
      noexpr="$2"
      yesstr="$3"
      nostr="$4"
      # shellcheck disable=SC2034
      messages_codeset="$5" # unused here, but kept as documentation

      # Display Yes / No ? prompt into locale
      cat <<EOF >&2
* WARNING: Directory '${DEFAULT_APPIMAGES_INSTALL_ALL_USERS_KDENLIVE_HOME}' found, but you are using 'single user' mode.
Do you wan't to continue and install kdenlive for current user (${USER}) only ? (${yesstr} / ${nostr})
EOF

      # Read answer
      read -r yn

      # Test answer
      case "$yn" in
        # match only work with the character class from the expression
        "${yesexpr##^}"|"${yesstr}")
          echoinfo 'Use --install-user'
          break
          ;;

        "${noexpr##^}"|"${nostr}")
          MODE='system'
          echoinfo 'Use --install-all-users'
          break
          ;;

        *)
          echoinfo "Waiting for: ${yesexpr##^} / ${yesstr}  or ${noexpr##^} / ${nostr}"
          ;;
      esac
    done
  fi

  if [ "${MODE}" = 'system' ] ; then
    SUDO_IF_NEEDED='sudo'
  else
    SUDO_IF_NEEDED=''
  fi
}

function compute_DEFAULT_APPIMAGES_INSTALL_HOME {
  if [ "${MODE}" = 'system' ] ; then
    echo "${DEFAULT_APPIMAGES_INSTALL_ALL_USERS_KDENLIVE_HOME}"
  else
    echo "${HOME}/Apps/kdenlive"
  fi
}

function get_extension {
  rev <<<"${1}" | cut -d '.' -f1 | rev
}

# Verify magic number for Linux binaries
function check_binary {
  local -r file="$1"

  local filerealpath=
  filerealpath="$( realpath "${file}" )"

  local magic=
  magic="$( hexdump -n 4  -v -e '/1 "%02X "' "${file}" )"

  if [ "${magic}" = '7F 45 4C 46 ' ] ; then
    ${SUDO_IF_NEEDED} chmod -v +x "${file}" || return $?
    echoinfo "'${file}' is ready: (${filerealpath})"
  else
    echofatal "'${file}' look suspicious...  (MAGIC='${magic}'): (${filerealpath})"
    return 100
  fi
}

function find_last_version_url {
  find_version_url | head -n 1
  # Expected result:
  #     https://files.kde.org/kdenlive/release/kdenlive-19.04.3-x86_64.appimage

  return 0
}

function do_curl {
  local -r curl_url="$1"
  local -a curl_params=()

  if [ "${DEBUG:-}" != 'true' ] ; then
    curl_params+=(-s) # --silent
  fi
  curl_params+=(-S) # --show-error

  (
    [ "${DEBUG:-}" = 'true' ] && set -x
    curl "${curl_params[@]}" "${curl_url}"
  ) || return $?
}

function list_versions_curl_and_filter {
 local _page__content_=
  _page__content_="$( do_curl "${DEFAULT_KDENLIVE_LIST_VERSIONS_URL}" )" || return $?

  (
    [ "${DEBUG:-}" = 'true' ] && set -x
    grep '<a[[:blank:]]href="' <<<"${_page__content_}" |
      grep '\[DIR\]' |
      sed -E -e 's,^.*<a[[:blank:]](.*)</a>.*$,\1,g' -e 's,^.*href="(.*)/".*$,\1,g' |
      sort --reverse
  ) || return $?
}

function list_appimages_curl_and_filter_entry {
  local _base_version_="$1"

  local _url_=
  _url_="$(
    # shellcheck disable=SC2059 # Don't use variables in the printf format string. Use printf '..%s..' "$foo".
    printf "${DEFAULT_KDENLIVE_LIST_APPIMAGES_URL_PATTERN}" "${_base_version_}"
    )"

  local _page_content_=
  _page_content_="$( do_curl "${_url_}" )" || return $?

  local _files_=
  _files_=$(
    [ "${DEBUG:-}" = 'true' ] && set -x
    grep '<a[[:blank:]]href="' <<<"${_page_content_}" |
      grep '\.AppImage' |
      sed -E -e 's,^.*<a[[:blank:]](.*)</a>.*$,\1,g' -e 's,^.*href="(.*)".*$,\1,g' |
      sort --reverse
  ) || return $?

  local _file_=
  while read -r _file_ ; do
    if [ -n "${_file_}" ] ; then
      if [ "${DEBUG:-}" = 'true' ] ; then
        echoinfo "FOUND: ${_file_}"
      fi
      echo "${_url_}${_file_}"
    fi
  done <<<"${_files_}"
}

function list_appimages_curl_and_filter {
  local _versions_=
  _versions_="$( list_versions_curl_and_filter )" || return $?
  local -r _versions_

  local _version_=
  while read -r _version_ ; do
    if [ -n "${_version_}" ] ; then
      if [ "${DEBUG:-}" = 'true' ] ; then
        echoinfo "FOUND: ${_version_}"
      fi
      list_appimages_curl_and_filter_entry "${_version_}" || return $?
    fi
  done <<<"${_versions_}"
}

function find_version_url {
  local _files_urls_=
  _files_urls_="$( list_appimages_curl_and_filter )" || return $?
  local -r _files_urls_

  local _file_url_=
  while read -r _file_url_ ; do
    if [ -n "${_file_url_}" ] ; then
      if [ "${DEBUG:-}" = 'true' ] ; then
        echoinfo "FOUND: ${_file_url_}"
      fi
      echo "${_file_url_}"
    fi
  done <<<"${_files_urls_}"

  # Expected result:
  #     https://files.kde.org/kdenlive/release/kdenlive-19.04.3-x86_64.appimage
  #     https://files.kde.org/kdenlive/release/kdenlive-19.04.2f-x86_64.appimage
  #     https://files.kde.org/kdenlive/release/kdenlive-19.04.2e-x86_64.appimage
  #     ...
  #     https://files.kde.org/kdenlive/release/kdenlive-17.08.3-x86_64.AppImage
  #     https://files.kde.org/kdenlive/release/Kdenlive-17.04.1b-x86_64.AppImage
  #     https://files.kde.org/kdenlive/release/Kdenlive-16.12.2-x86_64.AppImage
}

function create_menu_entry {
  local -r menuFilename="$1"
  local -r menuName="$2"
  local -r execFullPath="$3"
  local -r iconFullPath="$4"

  # See:
  # - https://linux.die.net/man/1/xdg-desktop-menu
  # - https://developer.gnome.org/menu-spec/

  cat <<EOF | ${SUDO_IF_NEEDED} tee "${menuFilename}" || return 1
[Desktop Entry]
Name=${menuName}
Exec=${execFullPath}
Comment=
Terminal=false
Icon=${iconFullPath}
Type=Application
Categories=AudioVideo
EOF

  cat <<EOF
File '${menuFilename}' created

-----------------------------
$( ls -la --color "${menuFilename}" )
-----------------------------
EOF
}

function download_and_install {
  local -r bin_filename="$1"
  local -r bin_url="$2"

  ${SUDO_IF_NEEDED} wget "${bin_url}" -O "${bin_filename}" || return $?
  check_binary "${bin_filename}" || return $?
  ${SUDO_IF_NEEDED} chmod +x "${bin_filename}" || return $?
}

function menu_install {
  local execFullPath="$1" # R/W

  local bin_basename=
  bin_basename="$( basename "${execFullPath}" )" || return $?

  local iconFullPath=
  if [ -n "${kdenlive_icon_url}" ] ; then
    if [ "${MODE}" = 'system' ] ; then
      iconFullPath="${COMMON_MENU_ICON_FILENAME_PREFIX}.$( get_extension "${kdenlive_icon_url}" )" || return $?
    else
      iconFullPath="${appimages_install_home}/${bin_basename}.$( get_extension "${kdenlive_icon_url}" )" || return $?
    fi

    ${SUDO_IF_NEEDED} wget "${kdenlive_icon_url}" -O "${iconFullPath}" || return $?
  fi

  local menuFilename=
  local menuName=

  if [ "${MODE}" = 'system' ] ; then
    ${SUDO_IF_NEEDED} ln -sfv "${execFullPath}" "${COMMON_APPIMAGE_LINK}" || return $?

    execFullPath="${COMMON_APPIMAGE_LINK}"
    menuFilename="${COMMON_DESKTOP}"
    menuName="${COMMON_MENU_ENTRY}"
  else
    create_folder_for_menu_if_needed "$( dirname "${menuFilename}" )" || return $?
    menuFilename="${appimages_install_home}/${bin_basename}.desktop"
    menuName="${bin_basename%.[aA]pp[iI]mage}"
  fi

  install_or_update_menu || return $?
}

function install_or_update_menu {
  # TODO:
  # https://gitlab.freedesktop.org/xdg/xdg-utils/issues/82 (should be fixed first !)
  # local icon_name='x-applications-kdenlive-appimage'

  create_menu_entry "${menuFilename}" "${menuName}" "${execFullPath}" "${iconFullPath}" || return $?

  # https://askubuntu.com/questions/277190/how-to-package-an-application-icon-properly
  # Assume icon is SVG
  # xdg-icon-resource install --novendor "${iconFullPath}" "${icon_name}"

  ${SUDO_IF_NEEDED} xdg-desktop-menu install --novendor --mode "${MODE}" "${menuFilename}" || return $?
}

function create_folder_for_menu_if_needed {
  local -r menuFolder="$1"

  if [ ! -d "${menuFolder}" ] ; then
    echowarn "Folder '${menuFolder}' not found."
    mkdir -v "${menuFolder}" || return $?
  fi

  if [ ! -d "${menuFolder}" ] ; then
    echowarn "Can't create '${menuFilename}': folder not found."
    return 100
  fi
}

function prerequisites {
  local -r install_home="$1"

  ${SUDO_IF_NEEDED} mkdir -vp "${install_home}" || return $?

  if command -v apt >/dev/null ; then # FIX improve package manager handling.
    # Install ladspa-sdk and swh-plugins packages
    # See: https://forum.kde.org/viewtopic.php?f=272&t=127270
    if [ "${param_skip_update}" != 'true' ] && [ -z "$(find /var/cache/apt/pkgcache.bin -mmin -60)" ] ; then
      ${SUDO_IF_NEEDED} apt update
    fi

    ${SUDO_IF_NEEDED} apt install 'ladspa-sdk' 'swh-plugins' || return $?
  else
    echowarn 'apt not found. See https://forum.kde.org/viewtopic.php?f=272&t=127270'
  fi
}

function download_and_install_if_needed {
  local -r appimages_install_home="${param_appimages_install_home:-$( compute_DEFAULT_APPIMAGES_INSTALL_HOME )}"
  local -r kdenlive_icon_url="${param_icon_url:-${DEFAULT_KDENLIVE_ICON_URL}}"

  prerequisites "${appimages_install_home}" || return $?

  echoinfo "Installation folder: '${appimages_install_home}'"

  local last_url=
  last_url="$( find_last_version_url )"
  if [ -z "${last_url}" ] ; then
    cat <<EOF >&2

*** Cannot compute URL for last version of kdenlive.

Try to using options: --list --debug

EOF
    exit 1
  fi
  echoinfo "Last version: '${last_url}'"

  local local_bin=
  local_bin="${appimages_install_home}/$( basename "${last_url}" )"
  if [ -f "${local_bin}" ] ; then
    if [ "${param_force_menu}" != 'true' ] && [ "${param_force_download}" != 'true' ] ; then
      cat <<EOF
File already there '${local_bin}'

You already have last version.

Use --force to reinstall or --force-menu to update menu.
EOF
      exit 0
    fi
  fi

  if [ ! -f "${local_bin}" ] || [ "${param_force_download}" = 'true' ] ; then
    download_and_install "${local_bin}" "${last_url}" || return $?
  fi

  menu_install "${local_bin}" || return $?
}

function main {
  local action='install'
  local param=
  local param_appimages_install_home=
  local param_force_download='false'
  local param_force_menu='false'
  local param_icon_url=
  local param_skip_update='false'

  set_install_mode '' || return $?

  while [ "$#" -gt 0 ] ; do
    param="$1"
    shift

    case "${param}" in
      '-h'|'--help')
        usage
        exit 0
        ;;

      '-l'|'--list'             ) action='--list' ;;
      '-i'|'--install'          ) action='--install' ;;
      '-U'|'--install-user'     ) action='--install-user' ;;
      '-A'|'--install-all-users') action='--install-all-users' ;;

      '--home') param_appimages_install_home="$( P_param "$@" )" && shift || return $? ;;
      '--icon') param_icon_url="$( P_param "$@" )" && shift || return $? ;;

      '-F'|'--force'    ) param_force_menu='true' ; param_force_download='true' ;;
      '--force-download') param_force_download='true' ;;
      '--force-menu'    ) param_force_menu='true' ;;

      '--skip-update'   ) param_skip_update='true' ;;
      '--debug'         ) DEBUG='true' ;;

      *)
        echofatal "Unknown option: '${param}'"
        usage >&2
        exit 1
        ;;
    esac
  done

  case "${action}" in
    '--list')
      find_version_url || return $?
      ;;

    '--install-user'|'--install-all-users')
      set_install_mode "${action}" || return $?
      fix_install_mode_if_needed || return $?
      download_and_install_if_needed || return $?
      ;;

    '--install'|*)
      set_install_mode '' || return $?
      fix_install_mode_if_needed || return $?
      download_and_install_if_needed || return $?
      ;;
  esac
}

main "$@" || exit $?
